// ====================================================================
// Copyright (c) 1997, 1998 The Apache Group.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer. 
//
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//
// 3. All advertising materials mentioning features or use of this
//    software must display the following acknowledgment:
//    "This product includes software developed by the Apache Group
//    for use in the Apache HTTP server project (http://www.apache.org/)."
//
// 4. The names "Apache Server" and "Apache Group" must not be used to
//    endorse or promote products derived from this software without
//    prior written permission.
//
// 5. Redistributions of any form whatsoever must retain the following
//    acknowledgment:
//    "This product includes software developed by the Apache Group
//    for use in the Apache HTTP server project (http://www.apache.org/)."
//
// THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
// EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
// ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
// ====================================================================
//
// This software consists of voluntary contributions made by many
// individuals on behalf of the Apache Group and was originally based
// on public domain software written at the National Center for
// Supercomputing Applications, University of Illinois, Urbana-Champaign.
// For more information on the Apache Group and the Apache HTTP server
// project, please see <http://www.apache.org/>.

// JServ - Serve up Java servlets
// by Alexei Kosut <akosut@apache.org>

// Parts are based on examples from  _Java in a Nutshell_ by David Flanagan:
// Written by David Flanagan.  Copyright (c) 1996 O'Reilly & Associates.
// You may study, use, modify, and distribute this example for any purpose.
// This example is provided WITHOUT WARRANTY either expressed or implied.

// JServConnection.java:
// - apache.jserv.JServConnection
// - apache.jserv.JServInputStream
// - apache.jserv.JServOutputStrem

package org.apache.jserv;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.OutputStreamWriter;
import java.io.DataInputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.InputStreamReader;

import java.net.Socket;
import java.net.InetAddress;

import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.util.TimeZone;

import java.text.SimpleDateFormat;
import java.text.ParseException;

import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpUtils;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * This class is the thread that handles all communications between
 * the Java VM and the web server.
 *
 * @author Alexei Kosut
 * @author Francis J. Lacoste
 **/
public class JServConnection extends Thread
implements HttpServletRequest, HttpServletResponse, JServSendError,
    JServDebug.DebugConstants
{
    private Socket client;
    private DataInputStream in;
    private OutputStream out;

    private Hashtable servletMgrTable;

    // Servlet setup stuff
    private JServServletManager mgr;
    private JServContext context;
    private Servlet servlet;
    private String servletname;
    private String hostname;

    private String auth;
    private String signal;

    private JServInputStream servlet_in;
    private JServOutputStream servlet_out;

    private BufferedReader servlet_reader;
    private boolean called_getInput;
    
    private PrintWriter servlet_writer;
    private boolean called_getOutput;

    // HTTP Stuff
    private Hashtable headers_in;
    private Hashtable headers_out;
    private Hashtable env_vars;
    private Cookie[] cookies_in;
    private Vector cookies_out;

    private int status;
    private String status_string;

    private Hashtable params;
    private boolean got_input;

    private boolean sent_header;

    //Session stuff
    JServSession session;
    String requestedSessionId;
    boolean idCameAsCookie;
    boolean idCameAsUrl;

    /**
     * Initalize the streams and starts the thread.
     */
    public JServConnection(Socket client_socket, Hashtable mgrTable) {
	this.client = client_socket;
	this.servletMgrTable = mgrTable;

	// Check that the request comes from localhost.  We check
	// 127.0.0.1 explicitly and a local address.

	// ik: the reason is because IP return addresses will use the same
	// interface they went through.  A connection to localhost will
	// appear to come from localhost while a connection to one of the
	// same host's other interfaces will come from the address of that
	// interface. (ref RFC1122)

        if (JServHandler.localhostcheck) {
	    if ( !client.getInetAddress().equals(client.getLocalAddress()) &&
		 !client.getInetAddress().getHostAddress().equals("127.0.0.1"))
		{
		    try {
			client.close();
		    } catch  (IOException e) {
		    }
		    return;
		}
        }

	try {
	    this.in = new DataInputStream ( new BufferedInputStream( client.getInputStream() ) );
	    this.out = new BufferedOutputStream( client.getOutputStream() );
	} catch(IOException e) {
	    try {
		client.close();
	    } catch(IOException e2) {
	    }
	    JServDebug.trace( "Exception while getting socket streams: ",
			      SERVICE_REQUEST );
	    JServDebug.trace( e );
	    return;
	}
	this.start();
    }

    // Provide the service.
    public void run() {
	JServDebug.trace( "Initializing servlet request", SERVICE_REQUEST );
	// Set up hash tables
	headers_in = new Hashtable();
	headers_out = new Hashtable();
	env_vars = new Hashtable();
	
	//cookies_in is initialized after the header are parsed.
	cookies_out = new Vector( 20 );

	// Set up HTTP defaults
	status = SC_OK;
	status_string = null;

	sent_header = false;
	got_input = false;
	called_getInput = false;
	called_getOutput = false;

	auth = null;
	signal = null;

	// Read in the data
	try {
	    JServDebug.trace( "Reading request data", SERVICE_REQUEST );
	    // Read data, and check to make sure it read ok
	    boolean readok = readData();

	    if (!readok) {
		sendError(SC_INTERNAL_SERVER_ERROR,
			  "Malformed data send to JServ");
		return;
	    }

	    if (!JServHandler.checkAuthString(auth)) {
		sendError(SC_FORBIDDEN,
			  "Illegal access to JServ: incorrect auth string");
		return;
	    }
	}
	catch(IOException e) {
	    try {
		client.close();
	    }
	    catch(IOException e2) {
	    }
	    return;
	}

	// override environemnt values
	env_vars.put("GATEWAY_INTERFACE", JServHandler.version );

        // Look for a signal 
        if (signal != null) {
 	    JServDebug.trace("received signal "+signal, SIGNAL);
            if (signal.equals("01"))  // SIGHUP
                JServHandler.signal_handler.restartServlets();
            else if (signal.equals("15")) // SIGTERM
                JServHandler.signal_handler.destroyServlets(); 
            return;
        }

	// Look for the servlet
	if (servletname == null) {
	    sendError(SC_NOT_FOUND, "Received empty servlet name");
	    return;
	}

	//Find the servlet manager that handles this host
	mgr = (JServServletManager) servletMgrTable.get(hostname);

        if (mgr == null) {
            mgr = (JServServletManager) servletMgrTable.get("_default_");
        }

	if (mgr == null) {
	    sendError(SC_INTERNAL_SERVER_ERROR,
		      "No servlet manager setup for host \"" +
		      hostname + "\".");
	    return;
	}
	
	JServDebug.trace( "Parsing cookies", SERVICE_REQUEST );
	//Ok parse the cookies
	cookies_in = JServUtils.parseCookieHeader( getHeader( "Cookie" ) );

	//Now to the session stuff
	requestedSessionId = 
	    JServServletManager.getUrlSessionId( getQueryString() );
	idCameAsUrl = requestedSessionId != null;
	String cookieSession = 
	    JServServletManager.getCookieSessionId( cookies_in );
	idCameAsCookie = cookieSession != null;
	
	//FIXME: What do we do if url and cookie don't have same id ?
	requestedSessionId = ( requestedSessionId == null ) ? cookieSession :
	    requestedSessionId;

	if ( requestedSessionId != null ) {
	    JServDebug.trace( "Request is in session " + requestedSessionId,
			      SERVICE_REQUEST );
	    JServSession s = (JServSession)mgr.getSession(requestedSessionId );
	    if ( requestedSessionId != null && s != null ) s.access();
	}

	//Let the mgr check for changes.
	mgr.checkReload( this );
	try {
	    context = mgr.loadServlet(servletname, this);
	} catch ( ServletException initError ) {
	    sendError(SC_INTERNAL_SERVER_ERROR,
		      "Initialization error while loading the servlet: " + initError.getMessage());
	}
	if (context == null  ||  context.servlet == null) {
	    sendError(SC_INTERNAL_SERVER_ERROR,
		      "An unknown error occured loading the servlet.");
	    return;
	}
	servlet = context.servlet;

	// Set up a read lock on the servlet. Note that anytime
	// we return, we need to make sure to unlock this. Otherwise,
	// we'll end up holding onto the lock forever. Oops.
	// This is done in the finally clause. 
	//The lock is acquired outside of the try block so that 
	//we do not unlock a lock that was never acquired.
	try {
	  context.lock.readLock();
	} catch ( InterruptedException stop ) {
	  JServDebug.trace( "Caught interrupted exception while waiting for "+
			    "servlet " + servletname + ": sending error.",
			    SERVICE_REQUEST );
	  sendError(SC_INTERNAL_SERVER_ERROR,
		    "InterruptedException while waiting for servlet lock.");
	  return;
	}

	try {
	    // Set up the servlet's I/O streams
	    servlet_in = new JServInputStream();
	    servlet_out = new JServOutputStream();

	    // Start up the servlet
	    try {
		JServDebug.trace( "Calling service()", SERVICE_REQUEST );
		servlet.service(this, this);
	    } catch(Exception e) {
		sendError(e);
		return;
	    } catch(Error e) {
		sendError(e);
		throw (Error) e.fillInStackTrace();
	    }

	    // Make sure we've send the HTTP header, even if no 
	    // entity data has been
	    sendHttpHeaders();

	    // All done; close the connection
	    try {
                // - KNOWN BUG - After the header has been sent to Apache
                // all logs written afterwords are lost.
                // - FIXME - This means that stuff logged after the headers
                // have been sent won't go out during this request, but will
                // go out during the next request using this virtual host.
                // Until we get the protocol changes, this is the cleanest
                // solution without buffering all the servlet output.
                //context.log_messages.removeAllElements();

		out.flush();
		client.close();
	    }
	    catch(IOException e) {
	    }
	    return;
	}
	finally {
	    //Clean up 
	    context.lock.readUnlock();
	}
    }

    //return the hostname
		public String getHostName()
		{
			return hostname;
		}
    // Read in a line from the input

    private String readHexLine() throws IOException {
	int len=0;
	byte hex[] = new byte[4];
	byte str[];

	try {
	    // Read four bytes from the input stream
	    in.readFully(hex);

	    // Convert them from hex to decimal
	    len = Character.digit((char)hex[0], 16);
	    len = len << 4;
	    len += Character.digit((char)hex[1], 16);
	    len = len << 4;
	    len += Character.digit((char)hex[2], 16);
	    len = len << 4;
	    len += Character.digit((char)hex[3], 16);
	    
 	    JServDebug.trace( "will read "+len+" bytes for this line", 
 			      REQUEST_DATA );

	    // Prevent NegativeSizeArray exception
	    if(len < 0)
	      throw new IOException("Malformed request");
	      
	    // Read len bytes from the input stream
	    str = new byte[len];
	    in.readFully(str);
	}
	catch  (NumberFormatException e) {
	    return null;
	}

	// Return str as an String
	return new String(str);
    }

    // Read in all the data
    private boolean readData() throws IOException {
      for (;;) {
	String line;
	try {
	  line = readHexLine();
	}
	catch(IOException e) {
	  return false;
	}

	    JServDebug.trace ("read " + line, REQUEST_DATA );
	    char id;

	    // Oops..
	    if   (line == null) {
		JServDebug.trace("Unexpected end of data.", REQUEST_DATA );
		return false;
	    }			
	    // Alldone
	    if    (line.length() == 0) {
		JServDebug.trace("All data read.", REQUEST_DATA);
		return true;
	    }

	    // Get the identifier from the first character
	    try {
		id = line.charAt(0);
		line = line.substring(1);
	    }
	    catch(StringIndexOutOfBoundsException e) {
 		JServDebug.trace( "No data to command: " + line.charAt(0),
 				  REQUEST_DATA );
		return false;
	    }

	    // All ids' take one or two pieces of data separated by a tab.
	    StringTokenizer tokens = new StringTokenizer(line, "\t");
	    String token1, token2;

	    try {
		token1 = tokens.nextToken();
	    }
	    catch(NoSuchElementException e) {
 		JServDebug.trace("Missing first piece of data for " + id,
 				 REQUEST_DATA );
		return false;
	    }

	    try {
		token2 = tokens.nextToken();
	    }
	    catch(NoSuchElementException e) {
		token2 = null;
	    }

	    // Switch, depending on what the id is
	    switch (id) {
	    case 'A':		
		// Auth data
 		JServDebug.trace("Authorization: " + token1, REQUEST_DATA );
		auth = token1;
		//Token2 is ignored
		break;
	    case 'C':
//		String dirname = token1;	//At present we don't need this.
		servletname = token2;
 		JServDebug.trace("Servlet: " + token2, REQUEST_DATA );
		break;
	    case 'S':
		hostname = token1;
		//Token2 is ingored
 		JServDebug.trace("Hostname: " + token1, REQUEST_DATA );
		break;
	    case 'E':		// Env variable
		env_vars.put(token1, (token2 != null) ? token2 : "");
 		JServDebug.trace("Env: " + token1 + "=" + token2, 
 				 REQUEST_DATA);
		break;
	    case 'H':		// Header
		headers_in.put(token1.toLowerCase(),
			       (token2 != null) ? token2 : "");
 		JServDebug.trace("Header: " + token1 + "=" + token2,
 				 REQUEST_DATA );
		break;
           case 's':            // Signal
                signal = token1;
 		JServDebug.trace("Signal: " + token1, REQUEST_DATA );
                break;
	    default:		// What the heck is this?
 		JServDebug.trace("Unknown: " + token1 + "=" + token2,
 				 REQUEST_DATA);
		return false;
	    }
	}
    }

    /**
     * Finds a status string from one of the standard
     * status code.
     * @param sc The status code to find a descriptive string.
     * @return A string describing this status code.
     */
    public static final String findStatusString(int sc) {
	switch (sc) {
	case SC_ACCEPTED:
	    return "Accepted";
	case SC_BAD_GATEWAY:
	    return "Bad Gateway";
	case SC_BAD_REQUEST:
	    return "Bad Request";
	case SC_CREATED:
	    return "Created";
	case SC_FORBIDDEN:
	    return "Forbidden";
	case SC_INTERNAL_SERVER_ERROR:
	    return "Internal Server Error";
	case SC_MOVED_PERMANENTLY:
	    return "Moved Permanently";
	case SC_MOVED_TEMPORARILY:
	    return "Moved Temporarily";
	case SC_NO_CONTENT:
	    return "No Content";
	case SC_NOT_FOUND:
	    return "Not Found";
	case SC_NOT_IMPLEMENTED:
	    return "Method Not Implemented";
	case SC_NOT_MODIFIED:
	    return "Not Modified";
	case SC_OK:
	    return "OK";
	case SC_SERVICE_UNAVAILABLE:
	    return "Service Temporarily Unavailable";
	case SC_UNAUTHORIZED:
	    return "Authorization Required";
	default:
	    return "Response";
	}
    }				
    
    // Send the HTTP headers, prepare to send response
    protected void sendHttpHeaders() {
	if (sent_header)
	    return;

	sent_header = true;

	//Use a PrintWriter around the socket out.
	PrintWriter printOut = new PrintWriter( this.out );

 	JServDebug.trace( "Sending response headers.", SERVICE_REQUEST );
	// Send logging data
        if (context != null) {
            StringBuffer temp = new StringBuffer(200); String logMsg;
            synchronized (context.log_messages) {
                for (Enumeration e = context.log_messages.elements();
                    e.hasMoreElements() ;) {
                    temp.setLength(0);
                    temp.append("Servlet-Log: ").append(e.nextElement());
                    logMsg = temp.toString();
                    JServDebug.trace( logMsg, RESPONSE_HEADERS );
                    printOut.print( logMsg + "\r\n");
                }
                context.log_messages.removeAllElements();
            }
        }
 
	// Send the status info

	if (status_string == null) {
	    status_string = findStatusString(status);
	}

 	String statusLine = "Status: " + status + " " + status_string;
 	printOut.print( statusLine + "\r\n" );
 	JServDebug.trace( statusLine, RESPONSE_HEADERS );

	if (headers_out != null) {
	    // Send the headers
	    for (Enumeration e = headers_out.keys(); e.hasMoreElements();) {
		String key = (String) e.nextElement();
		String hdr = key + ": " + headers_out.get(key);
		printOut.print( hdr + "\r\n");
		JServDebug.trace( hdr, RESPONSE_HEADERS );
	    }
	}

	//Send the cookies
	Enumeration cookies = cookies_out.elements();
	while ( cookies.hasMoreElements() ) {
	    Cookie cookie = (Cookie)cookies.nextElement();
	    String cookieHdr = "Set-Cookie: " + 
		JServUtils.encodeCookie( cookie );
	    printOut.print( cookieHdr + "\r\n" );
	    JServDebug.trace( cookieHdr, RESPONSE_HEADERS );
	}
	// Send a terminating blank line
	printOut.print("\r\n");
	//Flush the PrintWriter
	printOut.flush();
    }


    //---------------------------------------- Implementation of ServletRequest
    /**
     * Returns the size of the request entity data, or -1 if not
     * known. Same as the CGI variable CONTENT_LENGTH.  
     */
    public int getContentLength() {
	String lenstr = (String) env_vars.get("CONTENT_LENGTH");
	if (lenstr == null)
	  return -1;

	try {
	    return Integer.parseInt( lenstr );
	} catch(NumberFormatException e) {
	    return -1;
	}
    }

    /**
     * Returns the Internet Media Type of the request entity data, 
     * or null if not known. Same as the CGI variable
     * CONTENT_TYPE. 
     */
    public String getContentType() {
	return (String) env_vars.get("CONTENT_TYPE");
    } 

    /**
     * Returns the protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.  Same as the CGI variable SERVER_PROTOCOL.
     */
    public String getProtocol() {
	return (String) env_vars.get("SERVER_PROTOCOL");
    } 

    /**
     * Returns the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".  Different schemes have different
     * rules for constructing URLs, as noted in RFC 1738.  The URL used
     * to create a request may be reconstructed using this scheme, the
     * server name and port, and additional information such as URIs.
     */
    public String getScheme() {
	// Stronghold and Ben Laurie's Apache-SSL set the "HTTPS" environment
 	// variable in secure mode.
	String val = (String) env_vars.get("HTTPS");
	if (val != null) {
	    if (val.equalsIgnoreCase("on"))
		return "https";
	    else if (val.equalsIgnoreCase("off"))
		return "http";
	}			
	// FIXME: We just don't have this information available. We'll
	// look at the port, and return https if it's 443, but that's
	// not a real solution.
	if (getServerPort() == 443)
	    return "https";

	return "http";
    }

    /**
     * Returns the host name of the server that received the request.
     * Same as the CGI variable SERVER_NAME.
     */
    public String getServerName() {
	return (String) env_vars.get("SERVER_NAME");
    } 

    /**
     * Returns the port number on which this request was received.
     * Same as the CGI variable SERVER_PORT.
     */
    public int getServerPort() {
	String portstr = (String) env_vars.get("SERVER_PORT");
	Integer port;

	if      (portstr == null)
	    return -1;

	try {
	    port = new Integer(portstr);
	}
	catch(NumberFormatException e) {
	    return -1;
	}

	return port.intValue();
    }

    /**
     * Returns the IP address of the agent that sent the request.
     * Same as the CGI variable REMOTE_ADDR.
     */
    public String getRemoteAddr() {
	return (String) env_vars.get("REMOTE_ADDR");
    }

    /**
     * Returns the fully qualified host name of the agent that sent the
     * request. Same as the CGI variable REMOTE_HOST.
     */
    public String getRemoteHost() {
	return (String) env_vars.get("REMOTE_HOST");
    } 
    
    /**
     * Applies alias rules to the specified virtual path and returns
     * the corresponding real path, or null if the translation can not
     * be performed for any reason.  For example, an HTTP servlet would
     * resolve the path using the virtual docroot, if virtual hosting
     * is enabled, and with the default docroot otherwise.  Calling
     * this method with the string "/" as an argument returns the
     * document root.
     * @param path The virtual path to be translated to a real path.
     */
    public String getRealPath(String path) {
	// FIXME: Make this somehow talk to Apache, do a subrequest
	// and get the real filename. Until then, we just tack the path onto
	// the doc root and hope it's right. *sigh*

	// DOCUMENT_ROOT is not a standard CGI var, although Apache always
	// gives it. So we allow for it to be not present.
	String doc_root = (String) env_vars.get("DOCUMENT_ROOT");

	if     (doc_root == null)
	    return null;

	return doc_root + path;
    } 

    /**
     * Returns an input stream for reading binary data in the request body.
     * @exception IllegalStateException if getReader has been
     *        called on this same request.
     * @exception IOException n other I/O related errors.
     */
    public ServletInputStream getInputStream() throws IOException {
	if ( servlet_reader != null )
	    throw new IllegalStateException( "getReader() has already been "+
					     "called." );
	got_input = true;
	called_getInput = true;
	return servlet_in;
    }			
    
    
    // Parameter stuff
    private boolean parseParams() {
	// Have we already done it?
	if (params != null)
	    return false;

	try {
	    //Is this a post or a get
	    String method = getMethod();
	    if ( method.equals( "GET" ) ) {
		params = HttpUtils.parseQueryString(getQueryString());	    
	    } else if ( method.equals( "POST" ) ) {
		// Hmm... must have given the stream through getInputStream()
		if (got_input)
		    return true;
		
		got_input = true;
		params = HttpUtils.parsePostData(getContentLength(), 
						 servlet_in);
		if(params == null)
		  params = new Hashtable();

		return false;
	    } else {
	        //Unknown method
	        return true;
	    }
	} catch(IllegalArgumentException e) { 
	    return true;
	}
	return false;
    }

    /**
     * Returns a string containing the lone value of the specified
     * parameter, or null if the parameter does not exist. For example,
     * in an HTTP servlet this method would return the value of the
     * specified query string parameter. Servlet writers should use
     * this method only when they are sure that there is only one value
     * for the parameter.  If the parameter has (or could have)
     * multiple values, servlet writers should use
     * getParameterValues. If a multiple valued parameter name is
     * passed as an argument, the return value is implementation
     * dependent.
     * @param name the name of the parameter whose value is required.
     * @deprecated Please use getParameterValues
     */
    public String getParameter(String name) {
	if (parseParams())
	    return null;

	Object val = params.get(name);

	if     (val == null)
	    return null;

	if     (val instanceof String[]) {
	    // It's an array, return the first element
	    return ((String[])val)[0];
	} return (String) val;
    }

    /**
     * added for the <servlet> tag support - RZ.
     * provides the ability to add to the request object
     * the parameter of an embedded servlet.
     */
    public void setParameter(String name, String value)
    {
	//add the parameter in the hashtable, overrides any previous value
	parseParams();
	if(params != null)
	    {
		//remove any previous parameter with same name
		params.remove(name);

		//then put in the new parameter value
		params.put(name, value);
	    }
    }

    /**
     * Returns the values of the specified parameter for the request as
     * an array of strings, or null if the named parameter does not
     * exist. For example, in an HTTP servlet this method would return
     * the values of the specified query string or posted form as an
     * array of strings.
     * @param name the name of the parameter whose value is required.
     */
    public String[] getParameterValues(String name) {
	if (parseParams())
	    return null;

	Object val = params.get(name);

	if     (val == null)
	    return null;

	if     (val instanceof String) {
	    // It's a string, convert to an array and return
	    String va[] =
	    {(String) val};
	    return va;
	} return (String[]) val;
    }

    /**
     * Returns the parameter names for this request as an enumeration
     * of strings, or an empty enumeration if there are no parameters
     * or the input stream is empty.  The input stream would be empty
     * if all the data had been read from the stream returned by the
     * method getInputStream.
     */
    public Enumeration getParameterNames() {
	// an "empty Enumeration"... *sigh*
	if (parseParams())
	    return (new Vector()).elements();
	return params.keys();
    } 
    
    /**
     * Returns the value of the named attribute of the request, or
     * null if the attribute does not exist.  This method allows
     * access to request information not already provided by the other
     * methods in this interface.  Attribute names should follow the
     * same convention as package names. 
     * The following predefined attributes are provided.
     *
     * <TABLE BORDER>
     * <tr>
     *        <th>Attribute Name</th>
     *        <th>Attribute Type</th>
     *        <th>Description</th>
     *        </tr>
     *
     * <tr>
     *        <td VALIGN=TOP>javax.net.ssl.cipher_suite</td>
     *        <td VALIGN=TOP>string</td>
     *        <td>The string name of the SSL cipher suite in use, if the
     *                request was made using SSL</td>
     *        </tr>
     *
     * <tr>
     *        <td VALIGN=TOP>javax.net.ssl.peer_certificates</td>
     *        <td VALIGN=TOP>array of java.security.cert.X509Certificate</td>
     *        <td>The chain of X.509 certificates which authenticates the client.
     *                This is only available when SSL is used with client
     *                authentication is used.</td>
     *        </tr>
     *
     * <tr>
     *        <td VALIGN=TOP>javax.net.ssl.session</td>
     *        <td VALIGN=TOP>javax.net.ssl.SSLSession</td>
     *        <td>An SSL session object, if the request was made using SSL.</td>
     *        </tr>
     *
     * </TABLE>
     *
     * <BR>
     * <P>The package (and hence attribute) names beginning with java.*,
     * and javax.* are reserved for use by Javasoft. Similarly, com.sun.*
     * is reserved for use by Sun Microsystems.
     * 
     * <p><b>Note</b> The above attributes are not yet implemented by
     * JServ.
     * <p>On the other hand, attribute named "org.apache.jserv.&lt;variable&gt;"
     * returns the content of the environment (CGI) variable 
     * "&lt;variable&gt;".
     */
    public Object getAttribute(String name) {
	// We return "org.apache.jserv.<variable>" as the contents
	// of environment (CGI) variable "<variable>"
	if (!name.startsWith("org.apache.jserv."))
	    return null;

	// interface to get attribute names
	if (name.equals("org.apache.jserv.attribute_names"))
	    return env_vars.keys();

	return env_vars.get(name.substring("org.apache.jserv.".length()));
    }

    /**
     * Returns a buffered reader for reading text in the request body.
     * This translates character set encodings as appropriate.
     * @exception IllegalStateException if getOutputStream has been
     *        called on this same request.
     * @exception IOException on other I/O related errors.
     * @exception UnsupportedEncodingException if the character set encoding
     *  is unsupported, so the text can't be correctly decoded.
     */
    public BufferedReader getReader() throws IOException {
	if ( called_getInput ) {
	    throw new IllegalStateException("Already called getOutputStream");
	} else if ( servlet_reader == null ) {
	    //UnsupportedEncodingException is thrown not by
	    //getCharacterEncoding, which only parses the content-type header
	    //which specifies the encoding, but by the Reader constructor, 
	    //which possibly cannot support the encoding specified.
	    String encoding = 
		JServUtils.parseCharacterEncoding( getContentType() );
	    InputStreamReader reader =
		new InputStreamReader(servlet_in, encoding);
	    servlet_reader = new BufferedReader( reader );
	    got_input = true;
	}
	return servlet_reader;
    }
    
    //------------------------------------ Implementation of HttpServletRequest
    /**
     * Gets the array of cookies found in this request.
     * @return the array of cookies found in this request.
     */
    public Cookie[] getCookies() {
	return cookies_in;
    }
 

    /**
     * Gets the HTTP method (for example, GET, POST, PUT) with which
     * this request was made. Same as the CGI variable REQUEST_METHOD.
     * @return the HTTP method with which this request was made.
     */
    public String getMethod() {
	return (String) env_vars.get("REQUEST_METHOD");
    }

    /**
     * Gets this request's URI as a URL.
     * @return this request's URI as a URL.
     */
    public String getRequestURI() {
	// If the web server's version is available, use it
        String uri = (String) env_vars.get("REQUEST_URI");
        if (uri != null)
            return uri;

	if (getPathInfo() != null)
	    return getServletPath() + getPathInfo();
	else
	    return getServletPath();
    }

    /**
     * Gets the part of this request's URI that refers to the servlet
     * being invoked. Analogous to the CGI variable SCRIPT_NAME.
     * @return the servlet being invoked, as contained in this
     * request's URI.
     */
    public String getServletPath() {
	return (String) env_vars.get("SCRIPT_NAME");
    }

    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string. Same as the CGI variable PATH_INFO.
     *
     * @return the optional path information following the servlet
     * path, but before the query string, in this request's URI; null
     * if this request's URI contains no extra path information.
     */
    public String getPathInfo() {
	return (String) env_vars.get("PATH_INFO");
    }

    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string, and translates it to a real path.  Same as the CGI
     * variable PATH_TRANSLATED.
     *
     * @return extra path information translated to a real path or null
     * if no extra path information is in the request's URI.
     */
    public String getPathTranslated() {
	return (String) env_vars.get("PATH_TRANSLATED");
    }

    /**
     * Gets any query string that is part of the servlet URI.  Same as
     * the CGI variable QUERY_STRING.
     * @return query string that is part of this request's URI, or null
     * if it contains no query string.
     */
    public String getQueryString() {
	return (String) env_vars.get("QUERY_STRING");
    }

    /**
     * Gets the name of the user making this request.  The user name is
     * set with HTTP authentication.  Whether the user name will
     * continue to be sent with each subsequent communication is
     * browser-dependent.  Same as the CGI variable REMOTE_USER.
     *
     * @return the name of the user making this request, or null if not
     * known.
     */
    public String getRemoteUser() {
	return (String) env_vars.get("REMOTE_USER");
    }

    /**
     * Gets the authentication scheme of this request.  Same as the CGI
     * variable AUTH_TYPE.
     * 
     * @return this request's authentication scheme, or null if none.
     */
    public String getAuthType() {
	return (String) env_vars.get("AUTH_TYPE");
    }

    /**
     * Gets the value of the requested header field of this request.
     * The case of the header field name is ignored.
     * @param name the String containing the name of the requested
     * header field.
     * @return the value of the requested header field, or null if not
     * known.
     */
    public String getHeader(String name) {
	return (String) headers_in.get(name.toLowerCase());
    }

    /**
     * Gets the value of the specified integer header field of this
     * request.  The case of the header field name is ignored.  If the
     * header can't be converted to an integer, the method throws a
     * NumberFormatException.
     * @param name  the String containing the name of the requested
     * header field.
     * @return the value of the requested header field, or -1 if not
     * found.
     */
    public int getIntHeader(String name) {
	String hdrstr = (String) headers_in.get(name.toLowerCase());
	if (hdrstr == null)
	    return -1;
	
	return Integer.parseInt( hdrstr );
    }

    /**
     * Gets the value of the requested date header field of this
     * request.  If the header can't be converted to a date, the method
     * throws an IllegalArgumentException.  The case of the header
     * field name is ignored.
     *
     * <PRE>  From RFC2068:
     *  3.3.1 Full Date
     *
     *
     *   HTTP applications have historically allowed three different formats
     *   for the representation of date/time stamps:
     *
     *    Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
     *    Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
     *    Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
     *
     *   The first format is preferred as an Internet standard and
     *   represents a fixed-length subset of that defined by RFC 1123
     *   (an update to RFC 822).  The second format is in common use,
     *   but is based on the obsolete RFC 850 [12] date format and
     *   lacks a four-digit year.  HTTP/1.1 clients and servers that
     *   parse the date value MUST accept all three formats (for
     *   compatibility with HTTP/1.0), though they MUST only generate
     *   the RFC 1123 format for representing HTTP-date values in
     *   header fields
     * <pre>
     * @param name  the String containing the name of the requested
     * header field.
     * @return the value the requested date header field, or -1 if not
     * found.
     */
    public long getDateHeader(String name) {
	String val = (String) headers_in.get(name.toLowerCase());

	if ( val == null )
	    return -1;

	SimpleDateFormat sdf =
	    new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");
	try {
	    Date date = sdf.parse(val);
	    return date.getTime();
	}
	catch(ParseException e) {
	    ;			// Try another format

	}

	sdf = new SimpleDateFormat("EEEEEE, dd-MMM-yy HH:mm:ss zzz");
	try {
	    Date date = sdf.parse(val);
	    return date.getTime();
	}
	catch(ParseException e) {
	    // Try another format
	}

	sdf = new SimpleDateFormat("EEE MMMM d HH:mm:ss yyyy");
	try {
	    Date date = sdf.parse(val);
	    return date.getTime();
	} catch(ParseException e) {
	    throw new IllegalArgumentException( val );
	}
    }

    /**
     * Gets the header names for this request.
     * @return an enumeration of strings representing the header names
     * for this request. Some server implementations do not allow
     * headers to be accessed in this way, in which case this method
     * will return null.
     */
    public Enumeration getHeaderNames() {
	return headers_in.keys();
    }

    /**
     * Gets the current valid session associated with this request, if
     * create is false or, if necessary, creates a new session for the
     * request, if create is true.
     *
     * <p><b>Note</b>: to ensure the session is properly maintained,
     * the servlet developer must call this method (at least once)
     * before any output is written to the response.
     *
     * <p>Additionally, application-writers need to be aware that newly
     * created sessions (that is, sessions for which
     * <code>HttpSession.isNew</code> returns true) do not have any
     * application-specific state.
     * 
     * @return the session associated with this request or null if
     * create was false and no valid session is associated
     * with this request.
     */
    public HttpSession getSession( boolean create ) {
	//FIXME: What happen if someone calls invalidate on a new session
	//and calls this method again ? Should we create another one, or
	//return the invalid one ?
	if ( session != null ) return session;
	if ( requestedSessionId != null ) {
	    session = (JServSession)mgr.getSession( requestedSessionId );
	    if ( session != null ) return session;
	}
	if ( create == true ) {
	    session = mgr.createSession( this );
	    return session;
	}
	return null;
    }
    
    /**
     * Gets the session id specified with this request.  This may
     * differ from the actual session id.  For example, if the request
     * specified an id for an invalid session, then this will get a new
     * session with a new id.
     *
     * @return the session id specified by this request, or null if the
     * request did not specify a session id.
     * @see #isRequestedSessionIdValid()
     */
    public String getRequestedSessionId() {
	return requestedSessionId;
    }
    
    /**
     * Checks whether this request is associated with a session that
     * is valid in the current session context.  If it is not valid,
     * the requested session will never be returned from the
     * <code>getSession</code> method.
     * @return true if this request is assocated with a session that is
     * valid in the current session context.
     * @see #getRequestedSessionId()
     */
    public boolean isRequestedSessionIdValid() {
	if ( requestedSessionId == null ) return false;
	return mgr.getSession( requestedSessionId ) != null;
    }

    /**
     * Checks whether the session id specified by this request came in
     * as a cookie.  (The requested session may not be one returned by
     * the <code>getSession</code> method.)
     * @return true if the session id specified by this request came in
     * as a cookie; false otherwise.
     * @see #getSession( boolean )
     */
    public boolean isRequestedSessionIdFromCookie() {
	return idCameAsCookie;
    }

    /**
     * Checks whether the session id specified by this request came in
     * as part of the URL.  (The requested session may not be the one
     * returned by the <code>getSession</code> method.)
     * @return true if the session id specified by the request for this
     * session came in as part of the URL; false otherwise.
     * @see #getSession( boolean )
     */
    public boolean isRequestedSessionIdFromUrl() {
	return idCameAsUrl;
    }
    
    //--------------------------------------- Implementation of ServletResponse
    /**
     * Sets the content length for this response.
     * @param len the content length.
     */
    public void setContentLength(int len) {
	Integer length = new Integer(len);

	headers_out.put("Content-Length", length.toString());
    }

    /**
     * Sets the content type for this response.  This type may later
     * be implicitly modified by addition of properties such as the MIME
     * <em>charset=&lt;value&gt;</em> if the service finds it necessary,
     * and the appropriate media type property has not been set.  This
     * response property may only be assigned one time.
     * @param type the content's MIME type
     */
    public void setContentType(String type) {
	headers_out.put("Content-Type", type);
    }
    
    /**
     * Returns an output stream for writing binary response data.
     * @exception IllegalStateException if getWriter has been
     *      called on this same request.
     * @exception IOException if an I/O exception has occurred.
     * @see getWriter
     */
    public ServletOutputStream getOutputStream() throws IOException {
	if ( servlet_writer != null ) 
	    throw new IllegalStateException( "Already called getWriter" );
	called_getOutput = true;
	return servlet_out;
    }

    /**
     * Returns a print writer for writing formatted text responses.  The
     * MIME type of the response will be modified, if necessary, to reflect
     * the character encoding used, through the <em>charset=...</em>
     * property.  This means that the content type must be set before
     * calling this method.
     * @exception IllegalStateException if getOutputStream has been
     *        called on this same request.
     * @exception IOException on other errors.
     * @exception UnsupportedEncodingException if the character set encoding
     * @see getOutputStream
     * @see setContentType
     */
    public PrintWriter getWriter() throws IOException {
	if ( called_getOutput ) 
	    throw new IllegalStateException("Already called getOutputStream.");
	else if ( servlet_writer == null ) {
	    //UnsupportedEncodingException is thrown not by
	    //getCharacterEncoding, which only parses the content-type header
	    //which specifies the encoding, but by the Writer constructor, 
	    //which possibly cannot support the encoding specified.
	    OutputStreamWriter out = 
		new OutputStreamWriter( servlet_out, getCharacterEncoding() );
	    servlet_writer = new PrintWriter( out ); 
		
	}
	return servlet_writer;
    }

    /**
     * Returns the character set encoding used for this MIME body.
     * The character encoding is either the one specified in the
     * assigned content type, or one which the client understands.
     * If no content type has yet been assigned, it is implicitly
     * set to <em>text/plain</em>
     */
    public String getCharacterEncoding() {
	String contentType = (String)headers_out.get("Content-Type");
	if ( contentType == null ) {
	    contentType = "text/plain";
	    setContentType( contentType );
	}
	return JServUtils.parseCharacterEncoding( contentType );
    }
    
    //--------------------------------- Implementation of HttpServletResponse
    /**
     * Adds the specified cookie to the response.  It can be called
     * multiple times to set more than one cookie.
     * @param cookie the Cookie to return to the client
     */
    public void addCookie( Cookie cookie ) {
	cookies_out.addElement( cookie );
    }

    /**
     * Checks whether the response message header has a field with
     * the specified name.
     * @param name the header field name.
     * @return true if the response message header has a field with
     * the specified name; false otherwise.
     */
    public boolean containsHeader(String name) {
	return headers_out.contains(name.toLowerCase());
    }

    /**
     * Sets the status code and message for this response.  If the
     * field had already been set, the new value overwrites the
     * previous one.  The message is sent as the body of an HTML
     * page, which is returned to the user to describe the problem.
     * The page is sent with a default HTML header; the message
     * is enclosed in simple body tags (&lt;body&gt;&lt;/body&gt;).
     * @param sc the status code.
     * @param sm the status message.
     */
    public void setStatus(int sc, String sm) {
	status = sc;
	status_string = sm;
    }

    /**
     * 
     * Sets the status code for this response.  This method is used to
     * set the return status code when there is no error (for example,
     * for the status codes SC_OK or SC_MOVED_TEMPORARILY).  If there
     * is an error, the <code>sendError</code> method should be used
     * instead.
     * @param sc the status code
     * @see sendError
     */
    public void setStatus(int sc) {
	setStatus(sc, null);
    }

    /**
     * Adds a field to the response header with the given name and value.
     * If the field had already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be
     * used to test for the presence of a header before setting its
     * value.
     * @param name the name of the header field
     * @param value the header field's value
     * @see containsHeader
     */
    public void setHeader(String name, String value) {
        int offset_of_newline;
        // We need to make sure no newlines are present in the header:
        if ((offset_of_newline = value.indexOf((int)'\n')) > 0) {
            char msgAsArray[] = value.toCharArray();
            msgAsArray[offset_of_newline] = ' ';
            while ((offset_of_newline =
                    value.indexOf((int)'\n',offset_of_newline+1)) > 0) {
                 msgAsArray[offset_of_newline] = ' ';
            }
            value = new String(msgAsArray);
        }
	headers_out.put(name, value);
    }

    /**
     * Adds a field to the response header with the given name and
     * integer value.  If the field had already been set, the new value
     * overwrites the previous one.  The <code>containsHeader</code>
     * method can be used to test for the presence of a header before
     * setting its value.
     * @param name the name of the header field
     * @param value the header field's integer value
     * @see containsHeader
     */
    public void setIntHeader(String name, int value) {
	Integer val = new Integer(value);

	headers_out.put(name, val.toString());
    }

    /**
     * Adds a field to the response header with the given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * <code>containsHeader</code> method can be used to test for the
     * presence of a header before setting its value.
     * @param name the name of the header field
     * @param value the header field's date value
     * @see containsHeader
     */
    public void setDateHeader(String name, long date) {
	SimpleDateFormat sdf =
	    new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");
	TimeZone tz = TimeZone.getTimeZone("GMT");

	sdf.setTimeZone(tz);

	headers_out.put(name, sdf.format(new Date(date)));
    }

    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.  If setStatus has previously been
     * called, it is reset to the error status code.  The message is
     * sent as the body of an HTML page, which is returned to the user
     * to describe the problem.  The page is sent with a default HTML
     * header; the message is enclosed in simple body tags
     * (&lt;body&gt;&lt;/body&gt;).
     * @param sc the status code
     * @param msg the detail message
     */
    public void sendError(int sc, String msg) {
	try {
	    // Tell Apache to send an error
	    status = sc;
	    setHeader("Servlet-Error", msg);
	    sendHttpHeaders();

	    // Flush and close, so the error can be returned right
	    // away, and so any additional data sent is ignored
	    out.flush();
	    client.close();
	} catch (IOException e) {
	    // Not much more we can do...
	    JServDebug.trace( e );
	}
    }

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     * @param sc the status code
     */
    public void sendError(int sc) {
	sendError(sc, findStatusString( sc ) );
    }

    /**
     * JServSendError method. This sends an error message to Apache
     * when an exception occur in the ServletEngine. 
     */
    public void sendError(Throwable e) {
	sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
		  e.toString() + ": " + e.getMessage());
	JServDebug.trace( e );
    }

    /**
     * Sends a temporary redirect response to the client using the
     * specified redirect location URL.  The URL must be absolute (for
     * example, <code><em>https://hostname/path/file.html</em></code>).
     * Relative URLs are not permitted here.
     * @param location the redirect location URL
     * @exception IOException If an I/O error has occurred.
     */
    public void sendRedirect(String location) throws IOException {
	// We use Apache's internal status mechanism: set the status to
	// 200, with a Location: header and no body.

	setStatus(SC_OK);
	setHeader("Location", location);
	sendHttpHeaders();
    }

    /**
     * Encodes the specified URL by including the session ID in it,
     * or, if encoding is not needed, returns the URL unchanged.
     * The implementation of this method should include the logic to
     * determine whether the session ID needs to be encoded in the URL.
     * For example, if the browser supports cookies, or session
     * tracking is turned off, URL encoding is unnecessary.
     *
     * <p>All URLs emitted by a Servlet should be run through this
     * method.  Otherwise, URL rewriting cannot be used with browsers
     * which do not support cookies.
     * 
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     */
    public String encodeUrl( String url ) {
	//Encode only if there is no cookie support
	if ( idCameAsCookie ) return url;

	//And there is a valid session associoted with this request
	if ( session == null ) return url;
	
	return JServServletManager.encodeUrl( url, session.id );
    }

    /**
     * Encodes the specified URL for use in the
     * <code>sendRedirect</code> method or, if encoding is not needed,
     * returns the URL unchanged.  The implementation of this method
     * should include the logic to determine whether the session ID
     * needs to be encoded in the URL.  Because the rules for making
     * this determination differ from those used to decide whether to
     * encode a normal link, this method is seperate from the
     * <code>encodeUrl</code> method.
     *
     * <p>All URLs sent to the HttpServletResponse.sendRedirect
     * method should be run through this method.  Otherwise, URL
     * rewriting canont be used with browsers which do not support
     * cookies.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     * @see sendRedirect
     * @see encodeUrl
     */
    public String encodeRedirectUrl( String url ) {
	//Encode only if there is a session associated to the request
	if ( session == null ) return url;

	//And if the redirection will come back here
	if ( url.indexOf( hostname ) == -1 ) return url;

	return JServServletManager.encodeUrl( url, session.id );
    }

    //Stream are inner classes
    // ServletInputStream implementation
    private class JServInputStream extends ServletInputStream {

	//Default construcot
	public JServInputStream() {}

	// methods from InputStream
	public int read() throws IOException {
	    return in.read();
	}

	public int read(byte b[]) throws IOException {
	    return read(b, 0, b.length);
	}

	public int read(byte b[], int off, int len) throws IOException {
	    return in.read(b, off, len);
	}

	public long skip(long n) throws IOException {
	    return in.skip(n);
	}
  
	public void close() throws IOException {
	    //Ignore closing of the input stream since it also
	    //close the output stream.
	    //conn.in.close();
	}
    }

    // ServletOutputStream implementation
    class JServOutputStream extends ServletOutputStream {
    
	//Default constructor
	public JServOutputStream() {}
  
	// methods from OutputStream
	public void write(int b) throws IOException {
	    sendHttpHeaders();
	    out.write(b);
	}

	public void write(byte b[], int off, int len) throws IOException {
	    sendHttpHeaders();
	    out.write( b, off, len);
	}

	public void flush() throws IOException {
	    sendHttpHeaders();
	    out.flush();
	}

	public void close() throws IOException {
	    sendHttpHeaders();
	    out.close();
	}
    }
}
